/*
 * Copyright 2021 NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
 *
 * NOTICE TO LICENSEE:
 *
 * This source code and/or documentation ("Licensed Deliverables") are
 * subject to NVIDIA intellectual property rights under U.S. and
 * international Copyright laws.
 *
 * These Licensed Deliverables contained herein is PROPRIETARY and
 * CONFIDENTIAL to NVIDIA and is being provided under the terms and
 * conditions of a form of NVIDIA software license agreement by and
 * between NVIDIA and Licensee ("License Agreement") or electronically
 * accepted by Licensee.  Notwithstanding any terms or conditions to
 * the contrary in the License Agreement, reproduction or disclosure
 * of the Licensed Deliverables to any third party without the express
 * written consent of NVIDIA is prohibited.
 *
 * NOTWITHSTANDING ANY TERMS OR CONDITIONS TO THE CONTRARY IN THE
 * LICENSE AGREEMENT, NVIDIA MAKES NO REPRESENTATION ABOUT THE
 * SUITABILITY OF THESE LICENSED DELIVERABLES FOR ANY PURPOSE.  IT IS
 * PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND.
 * NVIDIA DISCLAIMS ALL WARRANTIES WITH REGARD TO THESE LICENSED
 * DELIVERABLES, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY,
 * NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * NOTWITHSTANDING ANY TERMS OR CONDITIONS TO THE CONTRARY IN THE
 * LICENSE AGREEMENT, IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY
 * SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THESE LICENSED DELIVERABLES.
 *
 * U.S. Government End Users.  These Licensed Deliverables are a
 * "commercial item" as that term is defined at 48 C.F.R. 2.101 (OCT
 * 1995), consisting of "commercial computer software" and "commercial
 * computer software documentation" as such terms are used in 48
 * C.F.R. 12.212 (SEPT 1995) and is provided to the U.S. Government
 * only as a commercial end item.  Consistent with 48 C.F.R.12.212 and
 * 48 C.F.R. 227.7202-1 through 227.7202-4 (JUNE 1995), all
 * U.S. Government End Users acquire the Licensed Deliverables with
 * only those rights set forth herein.
 *
 * Any use of the Licensed Deliverables in individual and commercial
 * software must include, in the user documentation and internal
 * comments to the code, the above Disclaimer and U.S. Government End
 * Users Notice.
 * </blockquote>}
 */


/**
 * @file
 * @brief This file contains all public function declarations of the cuTensorNet
 * library.
 */
#pragma once

#define CUTENSORNET_MAJOR 0 //!< cuTensorNet major version.
#define CUTENSORNET_MINOR 0 //!< cuTensorNet minor version.
#define CUTENSORNET_PATCH 1 //!< cuTensorNet patch version.
#define CUTENSORNET_VERSION (CUTENSORNET_MAJOR * 10000 + CUTENSORNET_MINOR * 100 + CUTENSORNET_PATCH)

#include <cuda_runtime.h>

#include <cutensornet/types.h>

#if defined(__cplusplus)
#include <cstdint>
#include <cstdio>

extern "C" {
#else
#include <stdint.h>
#include <stdio.h>

#endif /* __cplusplus */


/**
 * \brief Initializes the cuTensorNet library
 *
 * \details The device associated with a particular cuTensorNet handle is assumed to remain
 * unchanged after the cutensornetCreate() call. In order for the cuTensorNet library to 
 * use a different device, the application must set the new device to be used by
 * calling cudaSetDevice() and then create another cuTensorNet handle, which will
 * be associated with the new device, by calling cutensornetCreate().
 *
 * \param[out] handle Pointer to cutensornetHandle_t
 *
 * \returns CUTENSOR_STATUS_SUCCESS on success and an error code otherwise
 * \remarks blocking, no reentrant, and thread-safe
 */
cutensornetStatus_t cutensornetCreate(cutensornetHandle_t* handle);

/**
 * \brief Destroys the cuTensorNet library handle
 *
 * \details This function releases resources used by the cuTensorNet library handle. This function is the last call with a particular handle to the cuTensorNet library.
 * Calling any cuTensorNet function which uses cutensornetHandle_t after cutensornetDestroy() will return an error.
 *
 * \param[in,out] handle Opaque handle holding cuTensorNet's library context.
 */
cutensornetStatus_t cutensornetDestroy(cutensornetHandle_t handle);

/**
 * \mainpage cuTensorNet: A high-level CUDA library that is dedicated to operations on tensor networks (i.e., a collection of tensors) 
 */

/**
 * \brief Initializes a ::cutensornetNetworkDescriptor_t, describing connectivity between the tensors.
 *
 * Note that this function allocates data on the heap; hence, it is critical that cutensornetDestroyNetworkDescriptor() is called once \p descNet is no longer required.
 * 
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] numInputs Number of input tensors.
 * \param[in] numModesIn Array of size \p numInputs; ``numModesIn[i]`` denotes the number of modes available in the i-th tensor.
 * \param[in] extentsIn Array of size \p numInputs; ``extentsIn[i]`` has ``numModesIn[i]`` many entries with ``extentsIn[i][j]`` (``j`` < ``numModesIn[i]``) corresponding to the extent of the j-th mode of tensor ``i``.
 * \param[in] stridesIn Array of size \p numInputs; ``stridesIn[i]`` has ``numModesIn[i]`` many entries with ``stridesIn[i][j]`` (``j`` < ``numModesIn[i]``) corresponding to the linearized offset -- in physical memory -- between two logically-neighboring elements w.r.t the j-th mode of tensor ``i``.
 * \param[in] modesIn Array of size \p numInputs; ``modesIn[i]`` has ``numModesIn[i]`` many entries -- each entry corresponds to a mode. Each mode that does not appear in the input tensor is implicitly contracted.
 * \param[in] alignmentRequirementsIn Array of size \p numInputs; ``alignmentRequirementsIn[i]`` denotes the (minimal) alignment (in bytes) for the data pointer that corresponds to the i-th tensor (see ``rawDataIn[i]`` of cutensornetContraction()). It is recommended that each pointer is aligned to a 256-byte boundary.
 * \param[in] extentsOut Array of size \p numModesOut; ``extentsOut[j]`` (``j`` < ``numModesOut``) corresponding to the extent of the j-th mode of the output tensor.
 * \param[in] stridesOut Array of size \p numModesOut; ``stridesOut[j]`` (``j`` < ``numModesOut``) corresponding to the offset between two logically-neighboring elements within the j-th mode of the output tensor.
 * \param[in] numModesOut number of modes of the output tensor. On entry, if this value is 0 and the output modes are not provided, the network will infer the output modes.
 * \param[in] extentsOut Array of size \p numModesOut; ``extentsOut[j]`` denotes the extent of the j-th mode of the output tensor.
 * \param[in] stridesIn Array of size \p numOutputs; ``stridesOut[j]`` corresponding to the linearized offset -- in physical memory -- between two logically-neighboring elements w.r.t the j-th mode of the output tensor.
 * \param[in] modesOut Array of size \p numModesOut; ``modesOut[j]`` denotes the j-th mode of the output tensor.
 * \param[in] alignmentRequirementsOut Denotes the (minimal) alignment (in bytes) for the data pointer that corresponds to the output tensor (see rawDataOut of cutensornetContraction()). It's recommended that each pointer is aligned to a 256-byte boundary.
 * output tensor.
 * \param[in] dataType Denotes the data type for all input an output tensors.
 * \param[in] computeType Denotes the compute type used throughout the computation.
 * \param[out] descNet Pointer to a ::cutensornetNetworkDescriptor_t.
 *
 * \note If \p stridesIn (\p stridesOut) is set to 0 (\p NULL), it means the input tensors (output tensor) are in the Fortran (column-major) layout.
 *
 * Supported data-type combinations are:
 *
 * \verbatim embed:rst:leading-asterisk
 * +-------------+--------------------------+-------------+
 * |  Data type  |       Compute type       | Tensor Core |
 * +=============+==========================+=============+
 * | CUDA_R_16F  | CUTENSORNET_COMPUTE_32F  |   Volta+    |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_16BF | CUTENSORNET_COMPUTE_32F  |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_32F  | CUTENSORNET_COMPUTE_32F  |   No        |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_32F  | CUTENSORNET_COMPUTE_TF32 |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_32F  | CUTENSORNET_COMPUTE_16BF |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_32F  | CUTENSORNET_COMPUTE_16F  |   Volta+    |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_64F  | CUTENSORNET_COMPUTE_64F  |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_R_64F  | CUTENSORNET_COMPUTE_32F  |   No        |
 * +-------------+--------------------------+-------------+
 * | CUDA_C_32F  | CUTENSORNET_COMPUTE_32F  |   No        |
 * +-------------+--------------------------+-------------+
 * | CUDA_C_32F  | CUTENSORNET_COMPUTE_TF32 |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_C_64F  | CUTENSORNET_COMPUTE_64F  |   Ampere+   |
 * +-------------+--------------------------+-------------+
 * | CUDA_C_64F  | CUTENSORNET_COMPUTE_32F  |   No        |
 * +-------------+--------------------------+-------------+
 * \endverbatim

 */
cutensornetStatus_t cutensornetCreateNetworkDescriptor(const cutensornetHandle_t handle,
                                                  int32_t numInputs,
                                                  const int32_t numModesIn[],
                                                  const int64_t* const extentsIn[],
                                                  const int64_t* const stridesIn[],
                                                  const int32_t* const modesIn[],
                                                  const uint32_t alignmentRequirementsIn[],
                                                  int32_t numModesOut,
                                                  const int64_t extentsOut[],
                                                  const int64_t stridesOut[],
                                                  const int32_t modesOut[],
                                                  uint32_t alignmentRequirementsOut,
                                                  cudaDataType_t dataType,
                                                  cutensornetComputeType_t computeType,
                                                  cutensornetNetworkDescriptor_t* descNet);

/**
 * \brief Frees all the memory associated with desc
 *
 * \param[in,out] desc Opaque handle to a tensor network descriptor.
 */
cutensornetStatus_t cutensornetDestroyNetworkDescriptor(cutensornetNetworkDescriptor_t desc);

/**
 * \brief Computes the minimum workspace memory size needed to contract the input tensor-network using the provided contraction path.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] descNet Describes the tensor network (i.e., its tensors and their connectivity).
 * \param[in] optimizerInfo Opaque structure.
 * \param[out] workspaceSize The workspace memory size in bytes needed to successfully contract this tensor-network.
 */
cutensornetStatus_t cutensornetContractionGetWorkspaceSize(const cutensornetHandle_t handle,
                                                       const cutensornetNetworkDescriptor_t descNet,
                                                       const cutensornetContractionOptimizerInfo_t optimizerInfo,
                                                       uint64_t *workspaceSize);

/**
 * \brief Sets up the required hyper-optimization parameters for the contraction order solver (see cutensornetContractionOptimize())
 * 
 * Note that this function allocates data on the heap; hence, it is critical that cutensornetDestroyContractionOptimizerConfig() is called once \p optimizerConfig is no longer required.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[out] optimizerConfig This data structure holds all information about the user-requested hyper-optimization parameters.
 */
cutensornetStatus_t cutensornetCreateContractionOptimizerConfig(const cutensornetHandle_t handle,
                                                cutensornetContractionOptimizerConfig_t* optimizerConfig);

/**
 * \brief Frees all the memory associated with optimizerConfig
 *
 * \param[in,out] optimizerConfig Opaque structure.
 */
cutensornetStatus_t cutensornetDestroyContractionOptimizerConfig(cutensornetContractionOptimizerConfig_t optimizerConfig);

/**
 * \brief Get attributes of optimizerConfig.
 *
 * \param[in] handle Opaque handle holding cuTENSORNet's library context.
 * \param[in] optimizerConfig Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[out] buf On return, this buffer (of size sizeInBytes) holds the value that corresponds to attr within optimizerConfig.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionOptimizerConfigGetAttribute(const cutensornetHandle_t handle,
                                                cutensornetContractionOptimizerConfig_t optimizerConfig,
                                                cutensornetContractionOptimizerConfigAttributes_t attr,
                                                void *buf,
                                                size_t sizeInBytes);
/**
 * \brief Set attributes of optimizerConfig.
 *
 * \param[in] handle Opaque handle holding cuTENSORNet's library context.
 * \param[in] optimizerConfig Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[in] buf This buffer (of size sizeInBytes) determines the value to which attr will be set.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionOptimizerConfigSetAttribute(const cutensornetHandle_t handle,
                                                cutensornetContractionOptimizerConfig_t optimizerConfig,
                                                cutensornetContractionOptimizerConfigAttributes_t attr,
                                                const void *buf,
                                                size_t sizeInBytes);

/**
 * \brief Frees all the memory associated with optimizerInfo
 *
 * \param[in,out] optimizerInfo Opaque structure.
 */
cutensornetStatus_t cutensornetDestroyContractionOptimizerInfo(cutensornetContractionOptimizerInfo_t optimizerInfo);

/**
 * \brief Allocates resources for optimizerInfo.
 *
 * Note that this function allocates data on the heap; hence, it is critical that cutensornetDestroyContractionOptimizerInfo() is called once \p optimizerInfo is no longer required.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] descNet Describes the tensor network (i.e., its tensors and their connectivity) for which optimizerInfo is created.
 * \param[out] optimizerInfo Pointer to cutensornetContractionOptimizerInfo_t.
 */
cutensornetStatus_t cutensornetCreateContractionOptimizerInfo(const cutensornetHandle_t handle,
                                             const cutensornetNetworkDescriptor_t descNet,
                                             cutensornetContractionOptimizerInfo_t *optimizerInfo);

/**
 * \brief Computes an "optimized" contraction order as well as slicing info (for more information see Overview section) for a given tensor network.
 * 
 * Computes an "optimized" contraction order as well as slicing info (for more information see Overview section) for a given tensor network such that the total time to solution is minimized while adhering to the user-provided memory constraint.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] descNet Describes the topology of the tensor network (i.e., all tensors, their connectivity and modes).
 * \param[in] optimizerConfig Holds all hyper-optimization parameters that govern the search for an "optimal" contraction order.
 * \param[in] workspaceSizeConstraint Maximal device memory that will be provided by the user (i.e., cuTensorNet has to find a viable path/slicing solution within this user-defined constraint).
 * \param[in,out] optimizerInfo On return, this object will hold all necessary information about the optimized path and the related slicing information. optimizerInfo will hold information about (see cutensornetContractionOptimizerInfoAttributes_t):
 *      - Total number of slices.
 *      - Total number of sliced modes.
 *      - Information about the sliced modes (i.e., the ids of the sliced modes (see modesIn w.r.t. cutensornetCreateNetworkDescriptor()) as well as their extent (see Overview section for additional documentation).
 *      - Optimized path
 *      - Flop count
 *      - Size (in bytes) of the largest intermediate tensor.
 */
cutensornetStatus_t cutensornetContractionOptimize(const cutensornetHandle_t handle,
                                             const cutensornetNetworkDescriptor_t descNet,
                                             const cutensornetContractionOptimizerConfig_t optimizerConfig,
                                             uint64_t workspaceSizeConstraint,
                                             cutensornetContractionOptimizerInfo_t optimizerInfo);

/**
 * \brief Access attributes of optimizerInfo.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] optimizerInfo Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[out] buf On return, this buffer (of size sizeInBytes) holds the value that corresponds to attr within optimizeInfo.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionOptimizerInfoGetAttribute(
        const cutensornetHandle_t handle,
        const cutensornetContractionOptimizerInfo_t optimizerInfo,
        cutensornetContractionOptimizerInfoAttributes_t attr,
        void *buf,
        size_t sizeInBytes);

/**
 * \brief Sett attributes of optimizerInfo.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] optimizerInfo Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[in] buf This buffer (of size sizeInBytes) determines the value to which attr will be set.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionOptimizerInfoSetAttribute(
        const cutensornetHandle_t handle,
        cutensornetContractionOptimizerInfo_t optimizerInfo,
        cutensornetContractionOptimizerInfoAttributes_t attr,
        const void *buf,
        size_t sizeInBytes);

/**
 * \brief Initializes a cutensornetContractionPlan_t.
 *
 * Note that this function allocates data on the heap; hence, it is critical that cutensornetDestroyContractionPlan() is called once \p plan is no longer required.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] descNet Describes the tensor network (i.e., its tensors and their connectivity).
 * \param[in] optimizerInfo Opaque structure.
 * \param[in] workspaceSize Size of the provided workspace (in bytes).
 * \param[out] plan cuTensorNet's contraction plan holds all the information required to perform
 * the tensor contractions; to be precise, it initializes a cutensorContractionPlan for
 * each tensor contraction that is required to contract the entire tensor network.
*/
cutensornetStatus_t cutensornetCreateContractionPlan(const cutensornetHandle_t handle,
                                                     const cutensornetNetworkDescriptor_t descNet,
                                                     const cutensornetContractionOptimizerInfo_t optimizerInfo,
                                                     const uint64_t workspaceSize,
                                                     cutensornetContractionPlan_t* plan);

/**
 * \brief Frees all resources owned by plan.
 * 
 * \param[in,out] plan Opaque structure.
 */
cutensornetStatus_t cutensornetDestroyContractionPlan(cutensornetContractionPlan_t plan);

/**
 * \brief This function auto-tunes the contraction plan to find the best cutensorContractionPlan_t for each pair-wise contraction.
 *
 * This function auto-tunes the contraction plan to find the best cutensorContractionPlan_t for each pair-wise contraction.
 * This function is blocking due to the nature of the auto-tuning process.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in,out] plan The plan must already be created (see cutensornetCreateContractionPlan()); the individual contraction plans will be fine-tuned.
 * \param[in] rawDataIn Array of N pointers (N being the number of input tensors specified cutensornetCreateNetworkDescriptor()); rawDataIn[i] points to the data associated with the i-th input tensor (in device memory).
 * \param[out] rawDataOut Points to the raw data of the output tensor (in device memory).
 * \param[out] workspace Points to a scratchpad memory (in device memory).
 * \param[in] workspaceSize Size of the provided workspace (in bytes).
 * \param[in] pref Controls the auto-tuning process and gives the user control over how much time is spent in this routine.
 * \param[in,out] stream The CUDA stream in which all the computation is performed.
 * in 
 */
cutensornetStatus_t cutensornetContractionAutotune(const cutensornetHandle_t handle,
                                                   cutensornetContractionPlan_t plan,
                                                   const void* const rawDataIn[],
                                                   void* rawDataOut,
                                                   void* workspace, uint64_t workspaceSize,
                                                   const cutensornetContractionAutotunePreference_t pref,
                                                   cudaStream_t stream);

/**
 * \brief Sets up the required auto-tune parameters for the contraction plan
 * 
 * Note that this function allocates data on the heap; hence, it is critical that cutensornetDestroyContractionAutotunePreference() is called once \p autotunePreference is no longer required.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[out] autotunePreference This data structure holds all information about the user-requested auto-tune parameters.
 */
cutensornetStatus_t cutensornetCreateContractionAutotunePreference(const cutensornetHandle_t handle,
                                                cutensornetContractionAutotunePreference_t* autotunePreference);

/**
 * \brief Get attributes of AutotunePreference.
 *
 * \param[in] handle Opaque handle holding cuTENSORNet's library context.
 * \param[in] autotunePreference Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[out] buf On return, this buffer (of size sizeInBytes) holds the value that corresponds to attr within autotunePreference.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionAutotunePreferenceGetAttribute(const cutensornetHandle_t handle,
                                                cutensornetContractionAutotunePreference_t autotunePreference,
                                                cutensornetContractionAutotunePreferenceAttributes_t attr,
                                                void *buf,
                                                size_t sizeInBytes);
/**
 * \brief Set attributes of AutotunePreference.
 *
 * \param[in] handle Opaque handle holding cuTENSORNet's library context.
 * \param[in] autotunePreference Opaque structure that is accessed.
 * \param[in] attr Specifies the attribute that is requested.
 * \param[in] buf This buffer (of size sizeInBytes) determines the value to which attr will be set.
 * \param[in] sizeInBytes Size of buf (in bytes).
 */
cutensornetStatus_t cutensornetContractionAutotunePreferenceSetAttribute(const cutensornetHandle_t handle,
                                                cutensornetContractionAutotunePreference_t autotunePreference,
                                                cutensornetContractionAutotunePreferenceAttributes_t attr,
                                                const void *buf,
                                                size_t sizeInBytes);

/**
 * \brief Frees all the memory associated with autotunePreference
 *
 * \param[in,out] autotunePreference Opaque structure.
 */
cutensornetStatus_t cutensornetDestroyContractionAutotunePreference(cutensornetContractionAutotunePreference_t autotunePreference);

/**
 * \brief This function performs the actual contraction of the tensor network.
 *
 * \param[in] handle Opaque handle holding cuTensorNet's library context.
 * \param[in] plan Encodes the execution of a tensor network contraction (see cutensornetCreateContractionPlan and cutensornetContractionAutotune).
 * \param[in] rawDataIn Array of N pointers (N being the number of input tensors specified cutensornetCreateNetworkDescriptor()); rawDataIn[i] points to the data associated with the i-th input tensor (in device memory).
 * \param[out] rawDataOut Points to the raw data of the output tensor (in device memory).
 * \param[out] workspace Points to a scratchpad memory (in device memory).
 * \param[in] workspaceSize Size of the provided workspace (in bytes).
 * \param[in] sliceId The id of the slice that is currently contracted (this value ranges between 0 and optimizerInfo.numSlices); use 0 if no slices are used.
 * \param[in,out] stream The CUDA stream in which all the computation is performed.
 * in 
 */
cutensornetStatus_t cutensornetContraction(const cutensornetHandle_t handle,
                                           const cutensornetContractionPlan_t plan,
                                           const void* const rawDataIn[],
                                           void* rawDataOut,
                                           void* workspace, uint64_t workspaceSize,
                                           int64_t sliceId,
                                           cudaStream_t stream);

/**
 * \brief This function sets the logging callback routine.
 * \param[in] callback Pointer to a callback function. Check cutensornetLoggerCallback_t.
 */
cutensornetStatus_t cutensornetLoggerSetCallback(cutensornetLoggerCallback_t callback);

/**
 * \brief This function sets the logging output file.
 * \param[in] file An open file with write permission.
 */
cutensornetStatus_t cutensornetLoggerSetFile(FILE* file);

/**
 * \brief This function opens a logging output file in the given path.
 * \param[in] logFile Path to the logging output file.
 */
cutensornetStatus_t cutensornetLoggerOpenFile(const char* logFile);

/**
 * \brief This function sets the value of the logging level.
 * \param[in] level Log level, should be one of the following:
 * Level| Summary           | Long Description
 * -----|-------------------|-----------------
 *  "0" | Off               | logging is disabled (default)
 *  "1" | Errors            | only errors will be logged
 *  "2" | Performance Trace | API calls that launch CUDA kernels will log their parameters and important information
 *  "3" | Performance Hints | hints that can potentially improve the application's performance
 *  "4" | Heuristics Trace  | provides general information about the library execution, may contain details about heuristic status
 *  "5" | API Trace         | API Trace - API calls will log their parameter and important information
 */
cutensornetStatus_t cutensornetLoggerSetLevel(int32_t level);

/**
 * \brief This function sets the value of the log mask.
 *
 * \param[in]  mask  Value of the logging mask.
 * Masks are defined as a combination (bitwise OR) of the following masks:
 * Level| Description       |
 * -----|-------------------|
 *  "0" | Off               |
 *  "1" | Errors            |
 *  "2" | Performance Trace |
 *  "4" | Performance Hints |
 *  "8" | Heuristics Trace  |
 *  "16"| API Trace         |
 *
 * Refer to `cutensornetLoggerSetLevel` for the details.
 */
cutensornetStatus_t cutensornetLoggerSetMask(int32_t mask);

/**
 * \brief This function disables logging for the entire run.
 */
cutensornetStatus_t cutensornetLoggerForceDisable();

/**
 * \brief Returns Version number of the cuTensorNet library
 */
size_t cutensornetGetVersion();

/**
 * \brief Returns version number of the CUDA runtime that cuTensorNet was compiled against
 * \details Can be compared against the CUDA runtime version from cudaRuntimeGetVersion().
 */
size_t cutensornetGetCudartVersion();

/**
 * \brief Returns the description string for an error code
 * \param[in] error Error code to convert to string.
 * \returns the error string
 * \remarks non-blocking, no reentrant, and thread-safe
 */
const char *cutensornetGetErrorString(cutensornetStatus_t error);

#if defined(__cplusplus)
}
#endif /* __cplusplus */
